//package cypher;

/* part of the Cypher program, a java client for babel.
 *
 * this file deals with stuff coming in from the server
 *
 * functions in this file are as follows:
 *
 *   -  check_server(DataInputStream in) 
 *   -  dataReady(DataHeader hed)
 *   -  doPingReply()
 *   -  getLogin(DataHeader hed, byte[] bUser)
 *   -  getUserlist(DataHeader hed, byte[] uList)
 *   -  examineUser(DataHeader hed, byte[] bUser)
 *   -  showMessage(DataHeader hed, byte[] contents)
 *   -  showCookie(DataHeader hed, byte[] bCookie)
 *   -  getStatus(String sLetter)
 *   -  getPrivs(int iPrivs)
 *   -  replyAFK(int userID)
 */ 

import java.io.*;
import java.net.*;

public class Comms implements Constants
{
  private static String message; 
/*----------------------------------------------------------------------------}
{- quick function which checks the first DataHeader sent from the server contains
{- a C_HELLO
{----------------------------------------------------------------------------*/
public static int check_server(DataInputStream in) 
    {
      DataHeader login_head = new DataHeader();
      try {
        login_head.recv(in);
        if (login_head.getCommand() == C_HELLO) 
          System.out.println("Server acknowledgement. Please log in."); 
        else 
          System.out.println("received a \""+login_head.getCommand()+"\" from the server");
        return 0;
      }
      catch(Exception e) {
        System.err.println("Error receiving C_HELLO from the server.");
        return 1;
      }
    }

/*----------------------------------------------------------------------------}
{- dataReady is called each time a DataHeader is read from the server.
{- things should be received in the following order: C_HELLO C_LOGIN
{----------------------------------------------------------------------------*/
public static synchronized void dataReady(DataHeader hed, byte[] buffer)
    {
      // System.out.println("command - "+hed.getCommand()); // DEBUG
      // this is a huge set of 'if' statements which might be better handled by a case/switch statement?

      if (hed.getCommand() == C_HELLO) {
        //message = ("C_HELLO received");
        //System.out.println(message);
      }

      else if (hed.getCommand() == C_PING_REQ) {
        Cypher.DEBUG("C_PING_REQ received");
        doPingReply();
      }

      else if (hed.getCommand() == C_PING_REPLY) {
        Cypher.DEBUG("C_PING_REPLY received from server.");
        //System.out.println("Server replied to ping");
      }
      
      else if (hed.getCommand() == C_LOGIN) {
        getLogin(hed, buffer);     
      }

      else if (hed.getCommand() == C_LOGOUT) { 
        try {
          System.out.println("C_LOGOUT received");
          if (hed.getParam3()==P_LOGOUT_ZAP) {
            Cypher.DEBUG("You were zapped! P1:P2 - "+hed.getParam1()+":"+hed.getParam2());
          }
          Cypher.in.close();
          Cypher.out.close();
          System.out.println("streams have been closed. [cypher] session has ended.");
          System.out.println("<-- end the thread!!! -->");
          Input.logOut();
        }
        catch (IOException eClose) {
          message = ("error closing");
          System.out.println(message);
        }
      }

      else if (hed.getCommand() == C_USERLIST) {
        //message = ("C_USERLIST received");
        //System.out.println(message);
        getUserlist(hed, buffer);
      }
      
      // C_USER_CREATE goes here.

      else if (hed.getCommand() == C_USER_GET) {
        //message = ("C_USER_GET received");
        //System.out.println(message);
        examineUser(hed, buffer);
      }

      else if (hed.getCommand() == C_USER_PUT) {
        message = ("C_USER_PUT received");
        System.out.println(message);
        User xTarget = new User();
        xTarget = xTarget.translate(buffer);
        if (xTarget.Username.equals(Cypher.spod.Username)) {
          Cypher.spod.Popname = xTarget.Popname;
          System.out.println("Your popname was updated.");
        }else System.out.println("details of "+xTarget.Username+" received!");
      }

      else if (hed.getCommand() == C_MSG_GET) {
        showMessage(hed, buffer);
      }

      else if (hed.getCommand() == C_SET_COOKIE) {
        showCookie(hed, buffer);
      }

      else if (hed.getCommand() == C_WATCH_LOGIN) { // user logon. check to see if they're in our watch list.
        Cypher.DEBUG("C_WATCH_LOGIN received - "+hed.getParam1()+":"+hed.getParam2()+":"+hed.getParam3());
        getDetails(hed, buffer);
      }

      else if (hed.getCommand() == C_USERLIST_NEW) {
        getUserlist(hed, buffer);
      }

      else if (hed.getCommand() == C_WEBSERV_LIST) {
        Cypher.DEBUG("C_WEBSERV_LIST received - "+hed.getParam1()+":"+hed.getParam2()+":"+hed.getParam3());
        getDetails(hed, buffer); 
      }

      else {
        Cypher.DEBUG("Unknown header: "+hed.getCommand()+" - "+hed.getParam1()+":"+hed.getParam2()+":"+hed.getParam3());
        getDetails(hed, buffer);
      }
    } // end dataReady

  /*----------------------------------------------------------------------------}
  {- reply to a ping request from the server
  {----------------------------------------------------------------------------*/
  public static void doPingReply()
  {
    try {
      DataHeader pingHeader = new DataHeader(C_PING_REPLY, Cypher.spod.ConnectionNum, 0, 0, 0);
      pingHeader.send(Cypher.out);    	
    }
    catch (IOException ePingReply) {
      System.err.println(ePingReply);
      message = ("error whilst doing ping reply");
      System.out.println(message);
    }
  }


  /*----------------------------------------------------------------------------}
  {- server has sent back acceptance or denial of user login.
  {----------------------------------------------------------------------------*/
  public static void getLogin(DataHeader hed, byte[] bUser)
  {
    if (hed.getCommand()==C_LOGIN && hed.getParam3()==P_OK) {
      System.out.println("Login accepted.");
      message = ("Your connection number is "+hed.getParam1());
      System.out.println(message);
      //System.out.print("client status is " +hed.getParam2()+", ");
      //System.out.println("user status is " +hed.getParam3());
      Cypher.spod=Cypher.spod.translate(bUser);
      System.out.println("DEBUGGING: spod.popname is "+Cypher.spod.Popname);
      Cypher.spod.ConnectionNum = hed.getParam1();
      if ((Cypher.spod.Privs & S_INVIS_LOGIN)==S_INVIS_LOGIN) Cypher.spod.Invis=true;
      if ((Cypher.spod.Privs & S_ULTRAINVIS_LOGIN)==S_ULTRAINVIS_LOGIN) Cypher.spod.Ultra=true;
    }
    else {
      System.out.println("Login refused. Return param3 was "+hed.getParam3());
      if (hed.getParam3()==10){
        message = ("You are already logged in!");
        System.out.println(message);
      }
      if (hed.getParam3()==6){
        System.out.println("The server is not accepting any more connections.");
        System.out.println(message);
      }
      else {
        message = ("Well, that or your username/password were wrong.");
        System.out.println(message);
      }
    }
  } 


  /*----------------------------------------------------------------------------}
  {- updates the vector containg the userlist with the information coming in 
  {- from the server.
  {----------------------------------------------------------------------------*/
  public static void getUserlist(DataHeader hed, byte[] uList)
  {
    int numUsers=0;
    int x = 0;
    try {      
      numUsers=hed.getParam1();
      System.out.println("There are currently "+numUsers+" users on Babel."); //DEBUG
      
      Cypher.userList.clear();

      while(x < numUsers){
        NewSpod entry = new NewSpod();
        entry = entry.translate(uList, x);
        Cypher.userList.add(x, entry);
        x++;
      }

      /*
      for(int i=0;i<Cypher.userList.size();i++){
        NewSpod testSpod = new NewSpod();
        testSpod = (NewSpod)Cypher.userList.elementAt(i);
        //System.out.println(testSpod); //DEBUG
        //System.out.println(testSpod.Username+" - "+testSpod.ConID); //DEBUG
      }*/
      Input.showUserlist();
    }
    catch (Exception eUserlist) {
      message = ("Error storing Userlist: "+eUserlist);
      System.out.println(message);
    }
  }


  /*----------------------------------------------------------------------------}
  {- what gets returned from doing an Examine...
  {- what gets sent back if the user doesn't exist/misspelt name?
  {----------------------------------------------------------------------------*/
  public static void examineUser(DataHeader hed, byte[] bUser)
  {
    System.out.println("examining user....");
    User xTarget = new User();
    xTarget = xTarget.translate(bUser);
    System.out.println("Username: "+xTarget.Username);
    System.out.println("Popname: "+xTarget.Popname);
    System.out.println("Status: "+getStatus(xTarget.Status)); // you don't get status sent in NewSpod, only User.
    System.out.println("Real Name: "+xTarget.RealName);
    System.out.println("Priv Level: "+getPrivs(xTarget.Privs));// need a translation!
  }


  /*----------------------------------------------------------------------------}
  {- we've received a message, so tell us who it's from and display the message.
  {- can we get the username other than looping through the userlist?
  {- depends how it's stored, i guess.
  {----------------------------------------------------------------------------*/  
  public static void showMessage(DataHeader hed, byte[] contents)
  {
    String userName="";
    userName = Utils.IDToName(hed.getParam1());
    if (userName!=null) {
      //System.out.println("Message from "+userName);
      Cypher.DEBUG("Message received from: "+userName);
      Cypher.lastSender = hed.getParam1();
    }
    
    try 
    {
      DataInputStream mTemp = new DataInputStream(new ByteArrayInputStream(contents));
      String newMess = Utils.makeString(mTemp, hed.getPacketSize());
      message = ("Message was: "+newMess);
      System.out.println(message);
      if (Cypher.AFK == 1){ replyAFK(hed.getParam1());}// send AFK message.
    }
    catch (Exception eMakeString) {
      System.out.println("Error showing incoming message");
	  System.err.println(eMakeString);
    }
  } 





  /*----------------------------------------------------------------------------}
  {- we've received the cookie. tell the user what it is. (error check)
  {----------------------------------------------------------------------------*/
  public static void showCookie(DataHeader hed, byte[] bCookie)
  {
    try 
    {
      DataInputStream cTemp = new DataInputStream(new ByteArrayInputStream(bCookie));
      Cypher.spod.Cookie = Utils.makeString(cTemp, hed.getPacketSize());
      System.out.println("Cookie was: "+Cypher.spod.Cookie);
      Cypher.DEBUG("Got cookie: "+Cypher.spod.Cookie);
    }
    catch (Exception eMakeString)
    {
      System.err.println(eMakeString);
	  System.out.println("Error displaying cookie.");
    }
  } 

  /*----------------------------------------------------------------------------}
  {- translate the status letter into something sensible. :)
  {----------------------------------------------------------------------------*/
  private static String getStatus(String sLetter)
  {
    if (sLetter.equals("A")) {return "Active";}
    if (sLetter.equals("B")) {return "Banned";}
    if (sLetter.equals("D")) {return "Account Deleted";}
    if (sLetter.equals("S")) {return "Second Account";}
    if (sLetter.equals("P")) {return "Pending";}
    if (sLetter.equals("W")) {return "Webserver";}
    if (sLetter.equals("Q")) {return "Waiting to be registered? (queued)";}
    else return "Unknown Status field.";
  }


  /*----------------------------------------------------------------------------}
  {- translate the privs into something user-understandable.
  {----------------------------------------------------------------------------*/
  private static String getPrivs(int privs)
  {
	String privLevel = "";
    if (privs >= S_AUTHOR) {
		privLevel = "AUTHOR";
		privs = privs - S_AUTHOR - S_REGISTERED;
	} else if (privs >= S_BM) {
		privLevel = "SYSADMIN";
		privs = privs - S_BM - S_REGISTERED;
	} else if (privs >= S_SYSOP) {
		privLevel = "SYSOP";
		privs = privs - S_SYSOP - S_REGISTERED;
	} else if (privs >= S_REGISTERED) {
		privLevel = "REGISTERED USER";
		privs = privs - S_REGISTERED;	
	}
	
	if (privs >= S_ULTRAINVIS_LOGIN) {
		privLevel = privLevel+" - ULTRAINVIS Login";
		privs = privs - S_ULTRAINVIS_LOGIN;
	} 
	if (privs >= S_INVIS_LOGIN) {
		privLevel = privLevel+" - INVIS Login";
		privs = privs - S_INVIS_LOGIN;
	} 
    return privLevel;
  }

 /*----------------------------------------------------------------------------}
 {- we've received a message, but we're AFK, so reply to the user telling them so.
 {----------------------------------------------------------------------------*/
  private static void replyAFK(int userID)
  {
    char cUserID = (char)userID;
    String afkMessage = "Hi, this is Cypher, Pip's Java client. He's doing some work, and has turned off incoming messages, so won't see your message until he checks his log files. Sorry.";  
    String afkPacket=cUserID + afkMessage;
    int afkMessagesize=afkPacket.length();

    DataHeader afkHeader = new DataHeader(C_MSG_SEND,1,0,0,afkMessagesize);
    try
    {
      afkHeader.send(Cypher.out);
      Cypher.out.writeBytes(afkPacket);
      Cypher.out.flush();    	
    }
    catch (IOException eAFK)
    {
      message = ("Error whilst sending AFK message "+eAFK);
      System.out.println(message);
    }
  }

 /*----------------------------------------------------------------------------}
 {- we've received a header that isn't dealt with, so get the information and log it.
 {----------------------------------------------------------------------------*/
  private static void getDetails(DataHeader hed, byte[] buffer)
  {
    DataInputStream pPacket = new DataInputStream(new ByteArrayInputStream(buffer));
    try {
      Cypher.DEBUG(Utils.makeString(pPacket, hed.getPacketSize()));
      System.out.println(Utils.makeString(pPacket, hed.getPacketSize()));
    } catch (IOException eMakeString) {
      Cypher.DEBUG("Error converting packet data to String");
      Cypher.DEBUG(eMakeString);
    }
  }

}